@@ -1,10 +1,21 @@
 import re
 import os
-from typing import Any, Dict, List, Union
-from docassemble.base.functions import DANav
-from docassemble.base.util import log, word, DADict, DAList, DAObject, DAFile, DAFileCollection, DAFileList, defined, value, pdf_concatenate, zip_file, DAOrderedDict, action_button_html, include_docx_template, user_logged_in, user_info, action_argument, send_email, docx_concatenate, reconsider, get_config, space_to_underscore, LatitudeLongitude, DAStaticFile
+from typing import Any, Dict, List, Union, Callable
+from docassemble.base.util import log, word, DADict, DAList, DAObject, DAFile, DAFileCollection, DAFileList, defined, value, pdf_concatenate, zip_file, DAOrderedDict, action_button_html, include_docx_template, user_logged_in, user_info, send_email, docx_concatenate, get_config, space_to_underscore, DAStaticFile, alpha
 
-__all__ = ['ALAddendumField', 'ALAddendumFieldDict', 'ALDocumentBundle', 'ALDocument', 'ALStaticDocument', 'ALDocumentBundleDict','safeattr','label','key']
+__all__ = ['ALAddendumField',
+           'ALAddendumFieldDict',
+           'ALDocumentBundle', 
+           'ALDocument', 
+           'ALDocumentBundleDict',
+           'ALStaticDocument',
+           'safeattr',
+           'label',
+           'key',
+           'ALExhibitList',
+           'ALExhibit',
+           'ALExhibitDocument',
+           'unpack_dafilelist']
 
 DEBUG_MODE = get_config('debug')
 
@@ -403,94 +414,128 @@ class DALazyAttribute(DAObject):
           return dict()
 
 class ALDocument(DADict):
-  """
-  An opinionated dictionary of attachment blocks. Typically there are three:
-  1. The final version of a document with a signature. Use key="final" as an argument in a method.
-  2. The preview version of a document with no signature. Use key="preview" as an argument in a method.
-  3. An addendum of a document contained in the attribute `addendum`. E.g. `my_doc.addendum`.
+  """  
+  A dictionary of attachments, either created by a DAFile or an attachment
+  block. Typically there are three:
+  1. The final version of a document with a signature. E.g., my_doc['final'].
+  2. The preview version of a document with no signature. E.g.,
+     my_doc['preview'].
+  3. An addendum of a document contained in the attribute `addendum`. E.g.
+     `my_doc.addendum`.
 
-  Each form that an interview generates will get its own ALDocument object.
+  There is no limit to the number of keys, but the ALDocumentBundle class
+  expects at least a "final" key to exist, and the `addendum` attribute is
+  required if you desire to use PDF documents with text overflow. It is best
+  practice to use exactly the two keys "final" and "preview" and the attribute
+  "addendum". The "final" and "preview" documents will normally be the same
+  template, but with logic controlling the display of a particular section, such
+  as the signature.
 
-  This should really relate to one canonical document in different states. Not multiple
-  unrelated output documents that might get delivered together, except the addendum.
+  Each form that an interview generates should get its own ALDocument object.
 
-  The "addendum" attribute will typically be handled in a generic object block.
-  Multiple documents can use the same addendum template, with just the case caption
+  The "addendum" attribute can be handled in a generic object block. Multiple
+  documents can use the same addendum template, with just the case caption
   varying.
 
-  Required attributes:
-    - filename: name used for output PDF
-    - title: display name for the output PDF
-    - enabled: if this document should be created. See examples.
+  ALDocuments are designed to be used contingently as part of ALDocumentBundle
+  objects. Each ALDocument is considered to be "enabled" or "disabled" for a
+  particular interview user's session. This allows you to always work with a
+  single object representing all of the possible documents an interview can
+  generate, and use encapsulated logic to trigger individual documents inclusion
+  on the final download screen.
 
-  Optional attribute:
-    - addendum: an attachment block
-    - overflow_fields: ALAddendumFieldDict instance. These values will be used to detect and handle overflow.
-    - has_addendum: Defaults to False. Set to True if the document could have overflow, like for a PDF template.
+  Attributes:
+      filename (str): name used for output PDF
+      title (str): display name for the output PDF
+      enabled (bool): if this document should be created. See examples.
+      addendum (DAFile | DAFileCollection): (optional) an attachment block
+      overflow_fields (ALAddendumField): (optional) ALAddendumFieldDict
+        instance. These values will be used to detect and handle overflow.
+      has_addendum (bool): (optional) Defaults to False. Set to True if the
+        document could have overflow, like for a PDF template.
   
-  Examples:
+  Note:
+      The `enabled` attribute should always be defined by a code block or the
+      objects block, because by default it is considered fresh on each page
+      load. If your interview logic requires that you directly ask the user
+      whether or not to include a document, you can use a single intermediate
+      variable that is posed to the interview user to work around this
+      limitation.
+
+  Examples: # TODO: the code blocks aren't working right yet on the Docusaurus page.
+
+      Simple use where the document is always enabled and will have no addendum
+      --------------------------------------------------------------------------
+      ```yaml
+      ---
+      objects:
+        - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc", enabled=True)
+      --- 
+      attachment:
+        variable name: my_doc[i]  # This same template will be used for the `preview` and `final` keys
+        content: |
+          Here is some content
+
+          % if i == 'final':
+          ${ users[0].signature }
+          % elif i == 'preview':
+          [ Your signature here ]
+          % endif
+      ```
+
+      Enable a document conditionally
+      --------------------------------
+      ```yaml
+      ---
+      # See that `enabled` is not defined here
+      objects:
+        - affidavit_of_indigency: ALDocument.using(filename="affidavit-of-indigency.pdf", title="Affidavit of Indigency")
+      ---
+      code: |
+        affidavit_of_indigency.enabled = ask_indigency_questions and is_indigent
+      ```
+
+      An example enabling with a question posed to the interview user
+      ----------------------------------------------------------------
+      You should always use a code block or an object block to set the "enabled" status;
+      Use an intermediate variable if you want to ask the user directly whether or not to include a document.
+      ```yaml
+      ---
+      question: |
+        Do you want the extra document included?
+      yesno: include_extra_document
+      ---
+      code: |
+        extra_document.enabled = include_extra_document
+      ---
+      attachment:
+          variable name: extra_document[i] # This same template will be used for `final` and `preview`
+          docx template file: extra_document.docx
+      ```
   
-  Simple use where the document is always enabled and will have no addendum
-  ```
-  ---
-  objects:
-    - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc", enabled=True)
-  --- 
-  attachment:
-      variable name: my_doc[i]  # This will usually be "final" or "preview"
-      ...
-  ```
-  
-  Enable a document conditionally
-  ```
-  ---
-  # See that `enabled` is not defined here
-  objects:
-    - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc")
-    - affidavit_of_indigency: ALDocument.using(filename="affidavit-of-indigency.pdf", title="Affidavit of Indigency")
-  ---
-  # an example of an arbitrary condition
-  code: |
-    my_doc.enabled = condition1 and condition2
-    enable_my_doc_conditionally = True
-  ---
-  # A common example many interviews would encounter for an affidavit of indigency
-  code: |  
-    affidavit_of_indigency.enabled = ask_indigency_questions and is_indigent
-  ---
-  attachment:
-      variable name: my_doc[i]
-      ...
-  ---
-  attachment:
-      variable name: affidavit_of_indigency[i]
-      ...
-  ---
-  ```
-  
-  For a document that may need an addendum, you must specify this when the object is created
-  or in a mandatory code block. The addendum will only be triggered if the document has "overflow"
-  in one of the fields that you specify.
-  ```
-  ---
-  objects:
-    - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc", enabled=True, has_addendum=True)
-  --- 
-  attachment:
-      variable name: my_doc[i]
-      ...
-  ---
-  generic object: ALDocument
-  attachment:
-    variable name: x.addendum
-    docx template file: docx_addendum.docx
-  ---
-  code: |
-    my_doc.overflow_fields['big_text_variable'].overflow_trigger = 640 # Characters 
-    my_doc.overflow_fields['big_text_variable'].label = "Big text label" # Optional - you may use in your addendum
-    my_doc.overflow_fields['list_of_objects_variable'].overflow_trigger = 4 # Items in the list
-    my_doc.overflow_fields.gathered = True      
-  ```
+      For a document that may need an addendum, you must specify this when the object is created
+      or in a mandatory code block. The addendum will only be triggered if the document has "overflow"
+      in one of the fields that you specify.
+      ```
+      ---
+      objects:
+        - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc", enabled=True, has_addendum=True)
+      --- 
+      attachment:
+          variable name: my_doc[i]
+          ...
+      ---
+      generic object: ALDocument
+      attachment:
+        variable name: x.addendum
+        docx template file: docx_addendum.docx
+      ---
+      code: |
+        my_doc.overflow_fields['big_text_variable'].overflow_trigger = 640 # Characters 
+        my_doc.overflow_fields['big_text_variable'].label = "Big text label" # Optional - you may use in your addendum
+        my_doc.overflow_fields['list_of_objects_variable'].overflow_trigger = 4 # Items in the list
+        my_doc.overflow_fields.gathered = True      
+      ```
   """
   filename: str
   title: str
@@ -509,6 +554,8 @@ class ALDocument(DADict):
     if not hasattr(self, 'has_addendum'):
       self.has_addendum = False
     self.initializeAttribute('cache', DALazyAttribute)
+    self.always_enabled = hasattr(self, 'enabled') and self.enabled
+
 
   def as_pdf(self, key:str='final', refresh:bool=True) -> DAFile:
     # Trigger some stuff up front to avoid idempotency problems
@@ -529,7 +576,7 @@ class ALDocument(DADict):
     if refresh:
       main_doc = self.getitem_fresh(key)
     else:
-      main_doc = self.elements[key]
+      main_doc = self[key]
 
     if isinstance(main_doc, DAFileCollection):
       main_doc = main_doc.pdf
@@ -658,6 +705,9 @@ class ALStaticDocument(DAStaticFile):
     self.has_addendum = False
     self.auto_gather = False
     self.gathered = True
+    self.initializeAttribute('cache', DALazyAttribute)
+    self.always_enabled = hasattr(self, 'enabled') and self.enabled
+
   
   def __getitem__(self, key):
     # This overrides the .get() method so that the 'final' and 'private' key always exist and
@@ -729,6 +779,7 @@ class ALDocumentBundle(DAList):
     self.auto_gather=False
     self.gathered=True
     self.initializeAttribute('cache', DALazyAttribute)
+    self.always_enabled = hasattr(self, 'enabled') and self.enabled
 
   def as_pdf(self, key:str='final', refresh:bool=True) -> DAFile:
     safe_key = space_to_underscore(key)
@@ -743,7 +794,7 @@ class ALDocumentBundle(DAList):
       ending = ''
     else:
       ending = '.pdf'
-    files = self.enabled_documents()
+    files = self.enabled_documents(refresh=refresh)
     if len(files) == 1:
       # This case is simplest--we do not need to process the document at this level
       log_if_debug('Storing bundle for just one document ' + self.title + ' at ' + self.instanceName + '.cache.' + safe_key)
@@ -769,7 +820,7 @@ class ALDocumentBundle(DAList):
 
     # strip out a possible '.pdf' ending then add '.zip'
     zipname = os.path.splitext(self.filename)[0]
-    docs = [doc.as_pdf(key=key, refresh=refresh) for doc in self.enabled_documents()]
+    docs = [doc.as_pdf(key=key, refresh=refresh) for doc in self.enabled_documents(refresh=refresh)]
     zip = zip_file( docs, filename=zipname + '.zip' )
     if title == '':
       zip.title = self.title
@@ -784,11 +835,31 @@ class ALDocumentBundle(DAList):
   def preview(self, refresh:bool=True) -> DAFile:
     return self.as_pdf(key='preview', refresh=refresh)
 
-  def enabled_documents(self) -> List[Any]:
+  def enabled_documents(self, refresh:bool=True) -> List[Any]:
     """
     Returns the enabled documents
+    
+    Args:
+        refresh(bool): Controls whether the 'enabled' attribute is reconsidered.        
     """
-    return [document for document in self.elements if document.enabled]
+    if refresh:
+      retval = []
+      for document in self.elements:
+        if document.always_enabled:
+          enabled = True
+        else:
+          if hasattr(document.cache, 'enabled'):
+            enabled = document.cache.enabled
+          else:
+            document.cache.enabled = document.enabled
+            enabled = document.cache.enabled
+          if hasattr(document, 'enabled'):
+            del document.enabled
+        if enabled:
+          retval.append(document)
+      return retval
+    else:
+      return [document for document in self.elements if document.enabled]
 
   def as_flat_list(self, key:str='final', refresh:bool=True) -> List[DAFile]:
     """
@@ -798,7 +869,7 @@ class ALDocumentBundle(DAList):
     # Iterate through the list of self.templates
     # Unpack the list of documents at each step so this can be concatenated into a single list
     flat_list = []
-    for document in self.enabled_documents():
+    for document in self.enabled_documents(refresh=refresh):
       if isinstance(document, ALDocumentBundle):
         # call the bundle's as_flat_list() method to show all enabled templates.
         flat_list.extend(document.as_flat_list(key=key, refresh=refresh))
@@ -823,7 +894,7 @@ class ALDocumentBundle(DAList):
     """
     Returns the nested bundles as a list of PDFs that is only one level deep.
     """
-    return [doc.as_pdf(key=key, refresh=True) for doc in self if isinstance(doc, ALDocumentBundle) or doc.enabled]
+    return [doc.as_pdf(key=key, refresh=refresh) for doc in self.enabled_documents(refresh=refresh)]
 
   def as_editable_list(self, key:str='final', refresh:bool=True) -> List[DAFile]:
     """
@@ -836,59 +907,93 @@ class ALDocumentBundle(DAList):
       editable.append(doc.docx if hasattr(doc, 'docx') else doc.pdf)
     return editable
 
-  def download_list_html(self, key:str='final', format:str='pdf', view:bool=True, refresh:bool=True, include_zip:bool = True, view_label="View", view_icon:str="eye", download_label:str="Download", download_icon:str="download", zip_label:str="Download zip", zip_icon:str="file-archive") -> str:
+  def download_list_html(self, key:str='final', format:str='pdf', view:bool=True,
+      refresh:bool=True, include_zip:bool = True, view_label="View", view_icon:str="eye",
+      download_label:str="Download", download_icon:str="download", zip_label:str="Download zip",
+      zip_icon:str="file-archive") -> str:
     """
     Returns string of a table to display a list
     of pdfs with 'view' and 'download' buttons.
     """
     # Trigger some variables up top to avoid idempotency issues
-    for doc in self:
-      if doc.enabled:
-        doc.title
-        if format == 'pdf':
-          doc.as_pdf(key=key, refresh=refresh) # Generate cached file for this session
+    enabled_docs = self.enabled_documents(refresh=refresh)
+    for doc in enabled_docs:
+      doc.title
+      if format == 'pdf':
+        doc.as_pdf(key=key, refresh=refresh) # Generate cached file for this session
 
-    html ='<table class="al_table" id="' + html_safe_str(self.instanceName) + '">'
+    html = f'<table class="al_table" id="{ html_safe_str(self.instanceName) }">'
 
-    for doc in self:
+    for doc in enabled_docs:
       filename_root = os.path.splitext(str(doc.filename))[0]
-      if doc.enabled:
-        if format=='docx' and doc._is_docx(key=key):
-          download_doc = doc.as_docx(key=key)
-          download_filename = filename_root + ".docx"
-        else:
-          download_doc = doc.as_pdf(key=key)
-          download_filename = filename_root + ".pdf"
+      if format=='docx' and doc._is_docx(key=key):
+        download_doc = doc.as_docx(key=key)
+        download_filename = filename_root + ".docx"
+      else:
+        download_doc = doc.as_pdf(key=key)
+        download_filename = filename_root + ".pdf"
 
-        doc_download_button = action_button_html( download_doc.url_for(attachment=True, display_filename=download_filename), label=download_label, icon=download_icon, color="primary", size="md", classname='al_download' )
-        if view:
-          doc_view_button = action_button_html( doc.as_pdf(key=key).url_for(attachment=False, display_filename=filename_root + ".pdf"), label=view_label, icon=view_icon, color="secondary", size="md", classname='al_view' )
-          buttons = [ doc_view_button, doc_download_button ]
-        else:
-          buttons = [ doc_download_button ]
-        html += table_row( doc.title, buttons )
+      doc_download_button = action_button_html(
+          download_doc.url_for(attachment=True, display_filename=download_filename),
+          label=download_label, icon=download_icon, color="primary", size="md",
+          classname='al_download' )
+      if view:
+        doc_view_button = action_button_html(
+            doc.as_pdf(key=key).url_for(attachment=False, display_filename=filename_root + ".pdf"),
+            label=view_label, icon=view_icon, color="secondary", size="md", classname='al_view' )
+        buttons = [ doc_view_button, doc_download_button ]
+      else:
+        buttons = [ doc_download_button ]
+      html += table_row( doc.title, buttons )
     
     # Add a zip file row if there's more than one doc
     filename_root = os.path.splitext(str(self.filename))[0]
-    if len(self.enabled_documents()) > 1 and include_zip:
+    if len(enabled_docs) > 1 and include_zip:
       zip = self.as_zip(key=key)
-      zip_button = action_button_html( zip.url_for(attachment=False, display_filename = filename_root + ".zip"), label=zip_label, icon=zip_icon, color="primary", size="md", classname='al_zip' )
+      zip_button = action_button_html(
+          zip.url_for(attachment=False, display_filename = filename_root + ".zip"),
+          label=zip_label, icon=zip_icon, color="primary", size="md", classname='al_zip' )
       html += table_row( zip.title, zip_button)
-      
+
     html += '\n</table>'
 
     # Discuss: Do we want a table with the ability to have a merged pdf row?
     return html
 
   def download_html(self, key:str ='final', format:str ='pdf',
-                    view:bool=True, refresh:bool=True) -> str:
+                    view:bool=True, refresh:bool=True,
+                    view_label:str='View', view_icon:str='eye',
+                    download_label:str="Download", download_icon:str="download") -> str:
     """
     Returns an HTML string of a table to display all the docs
     combined into one pdf with 'view' and 'download' buttons.
     """
-    the_file = self.as_pdf(key=key)
+    if format == 'docx':
+      the_file = self.as_docx(key=key)
+    else:
+      the_file = self.as_pdf(key=key)
+
+    doc_download_button = action_button_html(
+        the_file.url_for(attachment=True),
+        label=download_label,
+        icon=download_icon,
+        color="primary",
+        size="md",
+        classname='al_download')
+    if view:
+      doc_view_button = action_button_html(
+          self.as_pdf(key=key).url_for(attachment=False),
+          label=view_label,
+          icon=view_icon,
+          color="secondary",
+          size="md",
+          classname='al_view')
+      buttons = [doc_view_button, doc_download_button]
+    else:
+      buttons = [doc_download_button]
+
     html ='<table class="al_table merged_docs" id="' + html_safe_str(self.instanceName) + '">'
-    html += table_row( self.title, the_file, download_file=the_file )
+    html += table_row(self.title, buttons)
     html += '\n</table>'
 
     return html
@@ -930,7 +1035,7 @@ class ALDocumentBundle(DAList):
   <div class="al_email_container">
     <span class="al_email_address '''+name+''' form-group row da-field-container da-field-container-datatype-email">
       <label for="'''+al_email_input_id+'''" class="al_doc_email col-form-label da-form-label datext-right">Email</label>
-      <input value="''' + (user_info().email if user_logged_in() else '') + '''" alt="Input box" class="form-control" type="email" size="35" name="'''+al_email_input_id+'''" id="'''+al_email_input_id+'''">
+      <input value="''' + (user_info().email if user_logged_in() else '') + '" alt="Input box" class="form-control" type="email" size="35" name="'+al_email_input_id+'" id="'+al_email_input_id+'''">
     </span>''' + action_button_html(javascript_string, label="Send", icon="envelope", color="primary", size="md", classname="al_send_email_button", id_tag=al_send_button_id) + "\n" + '''
   </div>
   '''
@@ -969,9 +1074,8 @@ class ALDocumentBundle(DAList):
 
 class ALDocumentBundleDict(DADict):
   """
-  A dictionary with named bundles of ALDocuments.
-  In the assembly line, we expect to find two predetermined bundles:
-  court_bundle and user_bundle.
+  A dictionary with named bundles of ALDocuments. In the assembly line, we
+  expect to find two predetermined bundles: court_bundle and user_bundle.
 
   It may be helpful in some circumstances to have a "bundle" of bundles. E.g.,
   you may want to present the user multiple combinations of documents for
@@ -999,3 +1103,252 @@ class ALDocumentBundleDict(DADict):
     Return a list of PDF-ified documents, suitable to make an attachment to send_mail.
     """
     return self[bundle].as_pdf_list(key='final')
+
+class ALExhibit(DAObject):
+  """Class to represent a single exhibit, with cover page, which may contain multiple documents representing pages.
+  Atributes:
+      elements (list): List of individual DAFiles representing uploaded images or documents.
+      cover_page (DAFile | DAFileCollection): (optional) A DAFile or DAFileCollection object created by an `attachment:` block
+        Will typically say something like "Exhibit 1"
+      label (str): A label, like "A" or "1" for this exhibit in the cover page and table of contents
+  """
+  cover_page: DAFile
+  pages: DAFileList
+  label: str
+  _cache: DALazyAttribute
+  starting_page: int
+
+  def init(self, *pargs, **kwargs):
+    super().init(*pargs, **kwargs)
+    self.initializeAttribute('_cache', DALazyAttribute)
+    self.object_type = DAFileList
+    if not hasattr(self, 'starting_page'):
+      self.start_page = 1
+
+  def add_numbers(self, starting_page=None)->None:
+    """
+    Todo:
+        Not implemented yet.
+    """
+    pass
+
+  def ocr_pages(self):
+    """
+    Return the OCR version if it exists; otherwise the initial version of each doc in `pages`.
+    """
+    pages = []
+    for page in self.pages:
+      if hasattr(page, 'ocr_version') and hasattr(page, 'ocr_status') and page.ocr_status.ready() and not page.ocr_status.failed() and page.ocr_version.ok:
+        pages.append(page.ocr_version)
+      else: 
+        pages.append(page)
+    return pages
+
+  def as_pdf(self, prefix='', add_page_numbers:bool=True, add_cover_page:bool=True, filename:str=None)->DAFile:
+    if hasattr(self._cache, '_file'):
+      return self._cache._file
+    if not filename:
+      filename = "exhibits.pdf"
+    if add_cover_page:
+      self._cache._file = pdf_concatenate(self.cover_page, self.ocr_pages(), filename=filename)
+    else:
+      self._cache._file = pdf_concatenate(self.ocr_pages(), filename=filename)
+    return self._cache._file
+
+  def num_pages(self)->int:
+    return self.pages.num_pages()
+
+  @property
+  def complete(self):
+    self.title
+    self.pages.gather()
+    return True
+
+  def __str__(self):
+    return self.title
+
+class ALExhibitList(DAList):
+  """
+  Attributes:
+      auto_label (bool): Set to True if you want exhibits to be automatically numbered for purposes of cover page 
+                         and table of contents. Defaults to True.
+      auto_labeler (Callable): (optional) a function or lambda to transform the index for each exhibit to a label.
+                               Defaults to labels like A..Z if unspecified.
+      auto_ocr (bool): Set to True if you would like exhibits to be OCR'ed in the background after they are uploaded. 
+                       Defaults to True.
+  """
+  def init(self, *pargs, **kwargs):
+    super().init(*pargs, **kwargs)
+    if not hasattr(self, 'auto_label'):
+      self.auto_label = True
+    if not hasattr(self, 'auto_labeler'):
+      self.auto_labeler = alpha
+    if not hasattr(self, 'auto_ocr'):
+      self.auto_ocr = True
+    self.object_type = ALExhibit
+    self.complete_attribute = 'complete'
+  
+  def as_pdf(self, filename="file.pdf", add_cover_pages:bool = True)->DAFile:
+    """
+    Return a single PDF containing all exhibits.
+    Args:
+        filename (str): the filename to be assigned to the generated PDF document.
+        add_cover_pages (bool): True if each exhibit should have a cover page, like "Exhibit A".
+    Returns:
+        A DAfile containing the rendered exhibit list as a single file.
+    """
+    return pdf_concatenate([exhibit.as_pdf(add_cover_page=add_cover_pages) for exhibit in self], filename=filename)
+
+  def add_numbers(self, prefix:str='', starting_number:int=1)->None:
+    """
+    Add running page numbers. (TODO: not yet implemented)
+    Args:
+        prefix (str): The prefix before each page number. E.g., Ex-
+        starting_number (int): The number that the first page will be assigned. Defaults 
+                               to 1.
+    """
+    pass
+
+  def _update_labels(self, auto_labeler:Callable=None)->None:
+    """
+    Private method to refresh labels on all exhibits.
+    Args:
+        auto_labeler (Callable): (optional) a lambda or function to transform index to a label, like A.
+    """
+    if auto_labeler is None:
+      auto_labeler = self.auto_labeler
+
+    for index, exhibit in enumerate(self.elements):
+      exhibit.label = self.auto_labeler(index)
+
+  def ocr_ready(self)->bool:
+    """
+    Returns:
+        True iff OCR process has finished on all pages. OCR is non-blocking, and assembly will work
+        even if OCR is not complete. Check this status if you want to wait to deliver a document until
+        OCR is complete.
+    """
+    ready = True
+    for exhibit in self.elements:
+      for page in exhibit.pages:
+        if hasattr(page, 'ocr_status'):
+          ready &= page.ocr_status.ready()
+    return ready
+
+  def _update_page_numbers(self)->None:
+    """
+    Update the `start_page` attribute of all exhibits so it reflects current position in the list + number of pages of each document.
+    """
+    current_index = 1
+    for exhibit in self.elements:
+      exhibit.start_page = current_index
+      current_index += exhibit.num_pages()
+
+  def _ocr_docs(self):
+    for exhibit in self.elements:
+      if len(exhibit.pages):
+        # We cannot OCR in place. It is too fragile.
+        for page in exhibit.pages:
+          page.ocr_version = DAFile(page.attr_name('ocr_version'))
+          # psm=1 is the default which uses automatic text orientation and location detection.
+          # Appears to be the most accurate method.
+          page.ocr_status = page.make_ocr_pdf_in_background(page.ocr_version, psm=1)
+
+  def hook_after_gather(self):
+    """
+    Private method automatically triggered when the list is fully gathered.
+    """
+    self._update_page_numbers()
+    if self.auto_label:
+      self._update_labels()
+    
+    # TODO: implement below. Was buggy when OCRing a lot of files and OCRing in place. 
+    # Switched to creating a new DAFile, but that introduced different bugs. -- third refresh breaks it. file never gets a number so not "ok"
+    # if self.auto_ocr:
+    #  self._ocr_docs()
+
+class ALExhibitDocument(ALDocument):
+  """Represents a collection of uploaded documents, formatted like a record appendix or exhibit list, with a table of contents and 
+  optional page numbering.
+
+  Attributes:
+      exhibits (ALExhibitList): list of ALExhibit documents. Each item is a separate exhibit, which may be multiple pages.
+      table_of_contents: DAFile or DAFileCollection object created by an `attachment:` block
+      _cache (DAFile): a cached version of the list of exhibits. It may take
+        a long time to process.
+      include_table_of_contents (bool): flag to control whether a table of contents is generated for this form        
+      include_exhibit_cover_pages (bool): flag to control whether cover pages are included with each separate exhibit
+      add_page_numbers (bool): Flag that controls whether the as_pdf() method
+        also will add Bates-stamp style page numbers and labels on each page.  
+
+  Todo:
+      * Method of making a safe link in place of the attachment (e.g., filesize limits on email)  
+      
+  Examples:
+  ```
+  ---
+  objects:
+    - exhibit_attachment: ALExhibitDocument.using(title="Exhibits", filename="exhibits" )
+  ---
+  code: |
+    exhibit_attachment.enabled = exhibit_attachment.exhibits.there_are_any
+  ---
+  objects:
+    - al_user_bundle: ALDocumentBundle.using(elements=[my_instructions, my_main_attachment, exhibit_attachment], filename="user_bundle.pdf", title="All forms to download for your records")
+  ```
+  """
+  exhibits: ALExhibitList
+  _cache: DAFile
+  table_of_contents: DAFile
+  include_table_of_contents: bool
+  include_exhibit_cover_pages: bool
+
+  def init(self, *pargs, **kwargs):
+    super().init(*pargs, **kwargs)
+    self.initializeAttribute('exhibits', ALExhibitList)
+    if not hasattr(self, 'include_table_of_contents'):
+      self.include_table_of_contents = True
+    if not hasattr(self, 'include_exhibit_cover_pages'):
+      self.include_exhibit_cover_pages = True
+    self.has_addendum = False
+
+  def has_overflow(self):
+    """
+    Provided for signature compatibility with ALDocument. Exhibits do not have overflow.
+    """
+    return False
+
+  def __getitem__(self, key):
+    # This overrides the .get() method so that the 'final' and 'private' key always exist and
+    # point to the same file.
+    return self
+  
+  def as_list(self, key:str='final', refresh:bool=True) -> List[DAFile]:
+    return [self]
+
+  def as_pdf(self, key="final", refresh:bool=True) -> DAFile:
+    """
+    Args:
+        key (str): unused, for signature compatibility with ALDocument
+    """
+    filename = os.path.splitext(self.filename)[0] + ".pdf"
+    
+    return pdf_concatenate(self.table_of_contents, self.exhibits.as_pdf(add_cover_pages=self.include_exhibit_cover_pages), filename=filename)
+    # pdf_concatenate([a.as_pdf() for a in self.exhibits], filename=self.filename)
+
+  def as_docx(self, key:str="bool", refresh:bool=True) -> DAFile:
+      return self.as_pdf()
+
+def unpack_dafilelist(the_file:DAFileList)->DAFile:
+  """Creates a plain DAFile out of the first item in a DAFileList
+  Args:
+      the_file (DAFileList): an item representing an uploaded document in a Docassemble interview
+  
+  Returns:
+      A DAFile representing the first item in the DAFileList, with a fixed instanceName attribute.
+  """
+  if isinstance(the_file, DAFileList):
+    temp_name = the_file.instanceName
+    the_file = next(iter(the_file))
+    the_file.instanceName = temp_name # reset instance name to the whole object instead of index in list we got rid of
+  return the_file
\ No newline at end of file
